#! /usr/bin/python
from __future__ import print_function
__version__ = "0.1"

import os, sys
import stat
import datetime
import subprocess
import codecs
import shutil
import time
import argparse # built in to Python 2.7 and 3.2
import json


myPath = os.path.dirname(os.path.realpath( __file__ )) # thanks to Jan Vorwerk's comment in http://code.activestate.com/recipes/474083-get-the-path-of-the-currently-executing-python-scr/

try:
	import hgapi
except:
	sys.path.append(os.path.join(myPath,'hgapi'))
	import hgapi

try:
	import hglib
except:
	sys.path.append(os.path.join(myPath,'python-hglib'))
	import hglib

try:
	import mwclient
except:
	sys.path.append(os.path.join(myPath,'mwclient'))
	import mwclient


def get_property(name, rev):
	if name + 'hidden' in rev:
		return '(hidden)'
	elif name == '*' and 'texthidden' in rev:
		return '(hidden)'
	else:
		return rev[name]

def prepend_line(path, line):
	if os.path.exists(path):
		with open(path, 'r') as old:
			os.unlink(path)
			with open(path, 'w') as new:
				new.write(str(line) + "\n")
				shutil.copyfileobj(old, new)
	else:
		with open(path, 'w') as new:
			new.write(str(line) + "\n")

def append_line(path, line):
	with open(path, 'a') as f:
		f.write(str(line) + "\n")

class HgRepoViaCommandLine():
	__repo = None
	newly_created = False
	path = None

	def __init__(self, repo_dir):
		"""Set up Mercurial repo."""
		self.newly_created = not os.path.isdir(repo_dir)
		if self.newly_created:
			os.mkdir(repo_dir)
		
		self.__repo = hgapi.Repo(repo_dir)
		if not os.path.isdir(os.path.join(repo_dir, '.hg')):
			self.__repo.hg_init()
		self.path = self.__repo.path

	def move_to_tip(self):
		self.__repo.hg_update('tip')
	
	def add(self, filenames):
		unknown_files = self.__repo.hg_status()['?']
		[self.__repo.hg_add(filename) for filename in unknown_files if filename in filenames ]

	def commit(self, message, user, timestamp):
		self.__repo.hg_commit(message=message, user=user, date=time.asctime(timestamp))

	def clean(self):
		return self.__repo.hg_status(empty=True) == {}

class HgRepoViaCommandServer():
	__repo = None
	newly_created = False
	path = None

	def __init__(self, repo_dir):
		"""Set up Mercurial repo."""
		self.newly_created = not os.path.isdir(repo_dir)
		if self.newly_created:
			os.mkdir(repo_dir)
		
		self.__repo = hgapi.Repo(repo_dir)
		if not os.path.isdir(os.path.join(repo_dir, '.hg')):
			self.__repo.hg_init()
		self.path = self.__repo.path

	def move_to_tip(self):
		self.__repo.hg_update('tip')
	
	def add(self, filenames):
		unknown_files = self.__repo.hg_status()['?']
		[self.__repo.hg_add(filename) for filename in unknown_files if filename in filenames ]

	def commit(self, message, user, timestamp):
		self.__repo.hg_commit(message=message, user=user, date=time.asctime(timestamp))

	def clean(self):
		return self.__repo.hg_status(empty=True) == {}


class HgRepo():
	@classmethod
	def vcs(cls):
		return "Mercurial"

	def __init__(self, repo_dir):
		self.imp = HgRepoViaCommandLine(repo_dir)

	def __getattr__(self, attr):
		return getattr(self.imp, attr)

def update_history(article, site, repo, start=None):
	# paths
	content_filename = article + '.wiki'
	revid_filename = 'revids.article'
	revid_talk_filename = 'revids.talk'
	content_path = os.path.join(repo.path, content_filename)
	revid_path = os.path.join(repo.path, revid_filename)

	# process start into something like '2011-01-01T00:00:00Z'
	if start:
		start = datetime.datetime.strptime(start, '%Y-%m-%d')
		start = start.isoformat()

	# get where we left off
	repo.move_to_tip()
	startid = None
	if os.path.isfile(revid_path):
		with open(revid_path, 'r') as f:
			for line in f:
				last_line = line
			startid = last_line.strip().split(' ')[0]
			startid = int(startid)

	# connect to mediawiki
	print("fetching from " + site + "...")
	site = mwclient.Site(site, clients_useragent='wikiarticle2hg/' + __version__)
	page = site.Pages[article]
	limit = None
	for rev in page.revisions(start=start, startid=startid, limit=limit, dir='newer', prop='ids|timestamp|flags|comment|user|content'):
		revid = rev['revid']
		if revid == startid:
			continue
		timestamp = rev['timestamp']
		content = get_property('*', rev)
		comment = get_property('comment', rev) or '(none)'
		user = get_property('user', rev) or '(none)'
		print("storing revision from", time.asctime(timestamp))
		
		with codecs.open(content_path, 'wb', encoding="utf-8") as f:
			f.write(content)

		append_line(revid_path, revid)

		repo.add([content_filename, revid_filename])
		repo.commit(message=comment, user=user, timestamp=timestamp)

def bzr_export():
	bzr_export = '/usr/lib/python2.6/dist-packages/bzrlib/plugins/fastimport/exporters/hg-fast-export.py'
	bzr_convert = "bzr fast-export-from-hg . - | bzr fast-import -"



def main():
	saved_args_filename = '.wiki2hg.args'

	repo_class = HgRepo
	repo_dir = None

	if len(sys.argv) == 1 and os.path.isfile(saved_args_filename):
		# try to read previous args from current directory
		with file(saved_args_filename) as f:
			args = json.load(f)
			repo_dir = os.getcwd()
	else:
		# args from command-line
		parser = argparse.ArgumentParser(description="Saves a Wikipedia article's history locally, within a VCS system, so it can be quickly searched through and analysed. {0} version.".format(repo_class.vcs()))
		parser.add_argument('articleName', metavar='"article name"', type=str)
		parser.add_argument('--site', metavar='site', default='en.wikipedia.org', help='Wiki site, defaults to en.wikipedia.org', type=str)
		parser.add_argument('--from', metavar='YYYY-MM-DD', help='date to start fetching revisions from', type=str, dest='startDate')
		args = vars(parser.parse_args())
		repo_dir = args['articleName']

	repo = repo_class(repo_dir)
	if not repo.clean():
		# can't be bothered handling cases where user has deleted our revids file, etc..
		print("please clean up the working copy: " + os.path.abspath(repo.path))
		sys.exit(1)
	else:
		# save args so that the repo can be updated without any command-line args
		saved_args_path = os.path.join(repo.path, saved_args_filename)
		if not os.path.exists(saved_args_path):
			with file(saved_args_path, 'w') as f:
				json.dump(args, f)
			repo.add(saved_args_filename)

		update_history(args['articleName'], args['site'], repo, args['startDate'])

		shellPath = os.environ.get('SHELL')
		comspec   = os.environ.get('COMSPEC')
		if repo.newly_created and shellPath and os.path.exists(shellPath) and os.getuid() > 0:
			# move user to new repository
			os.chdir(repo.path)
			print('')
			print("starting a shell in the repository's directory")
			os.execlp(shellPath, "") #FIXME: more appropriate function?
		elif repo.newly_created and comspec and comspec.endswith('\\cmd.exe') and sys.platform == "win32":
			# os.execlp("cmd.exe /K \"cd C:\my\destination && C:\"")
			# TODO
			pass


if __name__ == '__main__':
	main()
